/**
 * saltyrtc-task-relayed-data v0.3.1
 * A SaltyRTC Relayed Data task implementation.
 * https://github.com/saltyrtc/saltyrtc-task-relayed-data-js#readme
 *
 * Copyright (C) 2018 Threema GmbH
 *
 * This software may be modified and distributed under the terms
 * of the MIT license:
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
'use strict';

var saltyrtcTaskRelayedData = (function (exports) {
  'use strict';

  function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  }

  function _defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  function _createClass(Constructor, protoProps, staticProps) {
    if (protoProps) _defineProperties(Constructor.prototype, protoProps);
    if (staticProps) _defineProperties(Constructor, staticProps);
    return Constructor;
  }

  var RelayedDataTask =
  /*#__PURE__*/
  function () {
    function RelayedDataTask() {
      var debug = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : false;

      _classCallCheck(this, RelayedDataTask);

      this.debugging = false;
      this.initialized = false;
      this.eventRegistry = new saltyrtcClient.EventRegistry();
      this.debugging = debug;
    }

    _createClass(RelayedDataTask, [{
      key: "setDebug",
      value: function setDebug(enabled) {
        this.debugging = enabled;
      }
    }, {
      key: "debug",
      value: function debug() {
        if (this.debugging) {
          var _console;

          for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
          }

          (_console = console).debug.apply(_console, [this.logTag].concat(args));
        }
      }
    }, {
      key: "sendMessage",
      value: function sendMessage(data) {
        this.signaling.sendTaskMessage({
          type: RelayedDataTask.TYPE_DATA,
          p: data
        });
      }
    }, {
      key: "init",
      value: function init(signaling, data) {
        this.debug('init');
        this.signaling = signaling;
        this.initialized = true;
      }
    }, {
      key: "onPeerHandshakeDone",
      value: function onPeerHandshakeDone() {
        this.debug('onPeerHandshakeDone');
      }
    }, {
      key: "onTaskMessage",
      value: function onTaskMessage(message) {
        this.debug('New task message arrived: ' + message.type);

        switch (message.type) {
          case RelayedDataTask.TYPE_DATA:
            if (this.validateData(message) !== true) {
              return;
            }

            this.emit({
              type: RelayedDataTask.EVENT_DATA,
              data: message[RelayedDataTask.KEY_PAYLOAD]
            });
            break;

          default:
            console.error(this.logTag, 'Received message with invalid type:', message.type);
        }
      }
    }, {
      key: "sendSignalingMessage",
      value: function sendSignalingMessage(payload) {
        throw new saltyrtcClient.SignalingError(saltyrtcClient.CloseCode.ProtocolError, 'sendSignalingMessage called even though task does not implement handover');
      }
    }, {
      key: "getName",
      value: function getName() {
        return RelayedDataTask.PROTOCOL_NAME;
      }
    }, {
      key: "getSupportedMessageTypes",
      value: function getSupportedMessageTypes() {
        return [RelayedDataTask.TYPE_DATA];
      }
    }, {
      key: "getData",
      value: function getData() {
        return {};
      }
    }, {
      key: "close",
      value: function close(reason) {
        this.debug('Closing connection:', saltyrtcClient.explainCloseCode(reason));
      }
    }, {
      key: "on",
      value: function on(event, handler) {
        this.eventRegistry.register(event, handler);
      }
    }, {
      key: "once",
      value: function once(event, handler) {
        var _this = this;

        var onceHandler = function onceHandler(ev) {
          try {
            handler(ev);
          } catch (e) {
            _this.off(ev.type, onceHandler);

            throw e;
          }

          _this.off(ev.type, onceHandler);
        };

        this.eventRegistry.register(event, onceHandler);
      }
    }, {
      key: "off",
      value: function off(event, handler) {
        if (event === undefined) {
          this.eventRegistry.unregisterAll();
        } else {
          this.eventRegistry.unregister(event, handler);
        }
      }
    }, {
      key: "emit",
      value: function emit(event) {
        this.debug('New event:', event.type);
        var handlers = this.eventRegistry.get(event.type);
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = handlers[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var handler = _step.value;

            try {
              this.callHandler(handler, event);
            } catch (e) {
              console.error(this.logTag, 'Unhandled exception in', event.type, 'handler:', e);
            }
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator.return != null) {
              _iterator.return();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }
      }
    }, {
      key: "callHandler",
      value: function callHandler(handler, event) {
        var response = handler(event);

        if (response === false) {
          this.eventRegistry.unregister(event.type, handler);
        }
      }
    }, {
      key: "validateData",
      value: function validateData(message) {
        if (message[RelayedDataTask.KEY_PAYLOAD] === undefined) {
          console.warn(this.logTag, 'Data message does not contain payload');
          return false;
        }

        return true;
      }
    }, {
      key: "logTag",
      get: function get() {
        if (this.signaling === null || this.signaling === undefined) {
          return '[SaltyRTC.RelayedData]';
        }

        return '[SaltyRTC.RelayedData.' + this.signaling.role + ']';
      }
    }]);

    return RelayedDataTask;
  }();

  RelayedDataTask.PROTOCOL_NAME = 'v0.relayed-data.tasks.saltyrtc.org';
  RelayedDataTask.TYPE_DATA = 'data';
  RelayedDataTask.KEY_PAYLOAD = 'p';
  RelayedDataTask.EVENT_DATA = 'data';

  exports.RelayedDataTask = RelayedDataTask;

  return exports;

}({}));
